import os
import glob
import pandas as pd
import plotly.express as px


# Function to plot individual metric scores
def PlotMetrics(csv_path, method_path):
    metrics_df = pd.read_csv(csv_path)
    # Extract relevant metrics
    metrics = metrics_df.loc[
        metrics_df["Metrics"].isin(
            ["CSI_values", "POD_values", "Acc_values", "Prec_values", "F1_values"]
        )
    ].copy()

    metrics.loc[:, "Metrics"] = metrics["Metrics"].replace(
        {
            "CSI_values": "CSI",
            "POD_values": "POD",
            "Acc_values": "Accuracy",
            "Prec_values": "Precision",
            "F1_values": "F1 Score",
        }
    )
    value_columns = metrics.select_dtypes(include="number").columns

    for value_column in value_columns:
        metrics[value_column] = metrics[value_column].round(2)

        # Create the bar plot
        fig = px.bar(
            metrics,
            x=value_column,
            y="Metrics",
            title=f"Performance Metrics",
            labels={value_column: "Score"},
            text=value_column,
            color="Metrics",
            color_discrete_sequence=px.colors.qualitative.Set2,
        )
        fig.update_traces(texttemplate="%{text:.2f}", textposition="outside")
        fig.update_layout(
            yaxis_title="Metrics",
            xaxis_title="Score",
            showlegend=False,
            plot_bgcolor="rgba(0, 0, 0, 0)",
            paper_bgcolor="rgba(0, 0, 0, 0)",
            margin=dict(l=10, r=10, t=40, b=10),
            xaxis=dict(showline=True, linewidth=2, linecolor="black"),
            yaxis=dict(showline=True, linewidth=2, linecolor="black"),
            height=350,
            width=900,
            title_font=dict(family="Arial", size=24, color="black"),
            xaxis_title_font=dict(family="Arial", size=20, color="black"),
            yaxis_title_font=dict(family="Arial", size=20, color="black"),
            font=dict(family="Arial", size=18, color="black"),
        )

        # Save each plot as a PNG, using the column name as the filename
        plot_dir = os.path.join(method_path, "FinalPlots")
        if not os.path.exists(plot_dir):
            os.makedirs(plot_dir)

        output_filename = f"EvaluationMetrics_{value_column}.png"
        output_path = os.path.join(plot_dir, output_filename)

        # Save the plot as PNG
        fig.write_image(output_path, engine="kaleido", scale=500 / 96)
        print(
            f"Performance metrics chart ({value_column}) saved as PNG at {output_path}"
        )
        fig.show()


def PlotEvaluationMetrics(main_dir, method_name, out_dir):

    # If main directory contains the .tif files directly
    tif_files_main = glob.glob(os.path.join(main_dir, "*.tif"))
    if tif_files_main:
        method_path = os.path.join(out_dir, os.path.basename(main_dir), method_name)
        Evaluation_Metrics = os.path.join(method_path, "EvaluationMetrics")
        csv_files = os.path.join(Evaluation_Metrics, "EvaluationMetrics.csv")
        if not csv_files:
            print(f"No EvaluationMetrics CSV files found in '{Evaluation_Metrics}'.")
        else:
            PlotMetrics(csv_files, method_path)

    # Traverse all folders in main_dir if no .tif files directly in main_dir
    else:
        for folder in os.listdir(main_dir):
            folder_path = os.path.join(out_dir, folder)
            if os.path.isdir(folder_path):
                method_path = os.path.join(folder_path, method_name)
                Evaluation_Metrics = os.path.join(method_path, "EvaluationMetrics")
                csv_files = os.path.join(Evaluation_Metrics, "EvaluationMetrics.csv")
                if not csv_files:
                    print(
                        f"No EvaluationMetrics CSV files found in '{Evaluation_Metrics}'."
                    )
                else:
                    PlotMetrics(csv_files, method_path)



import shutil
import pyproj
import rasterio
from pathlib import Path
import geopandas as gpd
from rasterio.warp import calculate_default_transform, reproject, Resampling

#Lossless compression to reduce the file size
def compress_tif_lzw(tif_path):
    # Read original file
    with rasterio.open(tif_path) as src:
        profile = src.profile.copy()
        data = src.read()
    profile.update(compress='lzw')

    with rasterio.open(tif_path, 'w', **profile) as dst:
        dst.write(data)

#Check whether it is a projected CRS
def is_projected_crs(crs):
    return crs and crs.is_projected

#Check if the FIM bounds are within the CONUS
def is_within_conus(bounds, crs=None):
    CONUS_BBOX = (-125, 24, -66.5, 49.5) 
    left, bottom, right, top = bounds

    if crs and crs.is_projected:
        transformer = pyproj.Transformer.from_crs(crs, "EPSG:4326", always_xy=True)
        left, bottom = transformer.transform(left, bottom)
        right, top = transformer.transform(right, top)

    return (
        left >= CONUS_BBOX[0]
        and right <= CONUS_BBOX[2]
        and bottom >= CONUS_BBOX[1]
        and top <= CONUS_BBOX[3]
    )

#Reproject the FIMs to EPSG:5070 if withinUS and user doesnot define any target CRS, else user need to define it
def reprojectFIMs(src_path, dst_path, target_crs):
    with rasterio.open(src_path) as src:
        if src.crs != target_crs:
            transform, width, height = calculate_default_transform(
                src.crs, target_crs, src.width, src.height, *src.bounds
            )
            kwargs = src.meta.copy()
            kwargs.update({
                'crs': target_crs,
                'transform': transform,
                'width': width,
                'height': height
            })

            with rasterio.open(dst_path, 'w', **kwargs) as dst:
                for i in range(1, src.count + 1):
                    reproject(
                        source=rasterio.band(src, i),
                        destination=rasterio.band(dst, i),
                        src_transform=src.transform,
                        src_crs=src.crs,
                        dst_transform=transform,
                        dst_crs=target_crs,
                        resampling=Resampling.nearest
                    )
        else:
            print(f"Source raster is already in {target_crs}. No reprojection needed.")
            shutil.copy(src_path, dst_path)
        compress_tif_lzw(dst_path)

#Resample into the coarser resoution amoung all FIMS within the case
def resample_to_resolution(src_path, x_resolution, y_resolution):
    with rasterio.open(src_path) as src:
        transform = rasterio.transform.from_origin(src.bounds.left, src.bounds.top, x_resolution, y_resolution)
        width = int((src.bounds.right - src.bounds.left) / x_resolution)
        height = int((src.bounds.top - src.bounds.bottom) / y_resolution)

        kwargs = src.meta.copy()
        kwargs.update({
            'transform': transform,
            'width': width,
            'height': height
        })

        dst_path = src_path
        with rasterio.open(dst_path, 'w', **kwargs) as dst:
            for i in range(1, src.count + 1):
                reproject(
                    source=rasterio.band(src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src.crs,
                    dst_transform=transform,
                    dst_crs=src.crs,
                    resampling=Resampling.nearest
                )
    # compress_tif_lzw(dst_path)

#Check if the FIMs are in the same CRS or not else do further operation
def MakeFIMsUniform(fim_dir, target_crs=None, target_resolution=None):
    fim_dir = Path(fim_dir)
    tif_files = list(fim_dir.glob('*.tif'))
    if not tif_files:
        print(f"No TIFF files found in {fim_dir}")
        return
    processing_folder = fim_dir / 'processing'
    processing_folder.mkdir(exist_ok=True)

    crs_list = []
    projected_status = []
    bounds_list = []
    fims_data = []

    for tif_path in tif_files:
        try:
            with rasterio.open(tif_path) as src:
                crs_list.append(src.crs)
                projected_status.append(is_projected_crs(src.crs))
                bounds_list.append(src.bounds)
                fims_data.append((src.bounds, src.crs))
        except rasterio.RasterioIOError as e:
            print(f"Error opening {tif_path}: {e}")
            return

    all_projected = all(projected_status)
    first_crs = crs_list[0] if crs_list else None
    all_same_crs = all(crs == first_crs for crs in crs_list)

    if not all_projected or (all_projected and not all_same_crs):
        if target_crs:
            print(f"Reprojecting all FIMs to {target_crs}.")
            for src_path in tif_files:
                dst_path = processing_folder / src_path.name
                reprojectFIMs(str(src_path), str(dst_path), target_crs)
        else:
            all_within_conus = all(is_within_conus(bounds_list[i], crs_list[i]) for i in range(len(bounds_list)))

            if all_within_conus:
                default_target_crs = "EPSG:5070"
                print(f"FIMs are within CONUS, reprojecting all to {default_target_crs} and saving to {processing_folder}")
                for src_path in tif_files:
                    dst_path = processing_folder / src_path.name
                    reprojectFIMs(str(src_path), str(dst_path), default_target_crs)
            else:
                print("All flood maps are not in the projected CRS or are not in the same projected CRS.\n")
                print("Please provide a target CRS in EPSG format.")
    else:
        for src_path in tif_files:
            dst_path = processing_folder / src_path.name
            shutil.copy(src_path, dst_path)
            compress_tif_lzw(dst_path)
    
    # Resolution check and resampling
    processed_tifs = list(processing_folder.glob('*.tif'))
    if processed_tifs:
        resolutions = []
        for tif_path in processed_tifs:
            try:
                with rasterio.open(tif_path) as src:
                    resolutions.append(src.res)
            except rasterio.RasterioIOError as e:
                print(f"Error opening {tif_path} in processing folder: {e}")
                return

        first_resolution = resolutions[0] if resolutions else None
        all_same_resolution = all(res == first_resolution for res in resolutions)

        if not all_same_resolution:
            if target_resolution is not None:
                for src_path in processed_tifs:
                    resample_to_resolution(str(src_path), target_resolution, target_resolution)
            else:
                print("FIMs are in different resolution after projection. \n")
                coarser_x = max(res[0] for res in resolutions)
                coarser_y = max(res[1] for res in resolutions)
                print(f"Using coarser resolution: X={coarser_x}, Y={coarser_y}. Resampling all FIMS to this resolution.")
                for src_path in processed_tifs:
                    resample_to_resolution(str(src_path), coarser_x, coarser_y)
        else:
            print("All FIMs in the processing folder have the same resolution.")
    else:
        print("No TIFF files found in the processing folder after CRS standardization.")


import os
import ee
import glob
import geopandas as gpd
import rasterio
import pandas as pd
from pathlib import Path
from plotly.subplots import make_subplots
import plotly.graph_objects as go


def Changeintogpkg(input_path, output_dir, layer_name):
    input_path = str(input_path)
    # Check if the file is already in GPKG format
    if input_path.endswith(".gpkg"):
        return input_path
    else:
        # Convert to GPKG format if it's not
        gdf = gpd.read_file(input_path)
        output_gpkg = os.path.join(output_dir, f"{layer_name}.gpkg")
        gdf.to_file(output_gpkg, driver="GPKG")
        return output_gpkg


def Changeintogpkg(input_path, output_dir, layer_name):
    input_path = str(input_path)
    if input_path.endswith(".gpkg"):
        return input_path
    else:
        gdf = gpd.read_file(input_path)
        output_gpkg = os.path.join(output_dir, f"{layer_name}.gpkg")
        gdf.to_file(output_gpkg, driver="GPKG")
        return output_gpkg


def GetFloodedBuildingCountInfo(
    building_fp_path,
    study_area_path,
    raster1_path,
    raster2_path,
    contingency_map,
    save_dir,
    basename,
):
    output_dir = os.path.dirname(building_fp_path)

    building_fp_gpkg = Changeintogpkg(
        building_fp_path, output_dir, "building_footprint"
    )

    building_gdf = gpd.read_file(building_fp_gpkg)
    study_area_gdf = gpd.read_file(study_area_path)

    if building_gdf.crs != study_area_gdf.crs:
        building_gdf = building_gdf.to_crs(study_area_gdf.crs)

    clipped_buildings = gpd.overlay(building_gdf, study_area_gdf, how="intersection")
    clipped_buildings["centroid"] = clipped_buildings.geometry.centroid

    centroid_counts = {
        "Benchmark": 0,
        "Candidate": 0,
        "False Positive": 0,
        "False Negative": 0,
        "True Positive": 0,
    }

    def count_centroids_in_raster(raster_path, label):
        with rasterio.open(raster_path) as src:
            raster_data = src.read(1)
            transform = src.transform

            for centroid in clipped_buildings["centroid"]:
                row, col = src.index(centroid.x, centroid.y)
                if 0 <= row < raster_data.shape[0] and 0 <= col < raster_data.shape[1]:
                    pixel_value = raster_data[row, col]
                    if label in ["Benchmark", "Candidate"]:
                        if pixel_value == 2:  # False Positive
                            centroid_counts[label] += 1
                    else:
                        if pixel_value == 2:
                            centroid_counts["False Positive"] += 1
                        elif pixel_value == 3:
                            centroid_counts["False Negative"] += 1
                        elif pixel_value == 4:
                            centroid_counts["True Positive"] += 1

    if "benchmark" in str(raster1_path).lower():
        count_centroids_in_raster(raster1_path, "Benchmark")
        count_centroids_in_raster(raster2_path, "Candidate")
    elif "candidate" in str(raster2_path).lower():
        count_centroids_in_raster(raster1_path, "Candidate")
        count_centroids_in_raster(raster2_path, "Benchmark")

    if "contingency" in str(contingency_map).lower():
        count_centroids_in_raster(contingency_map, "Contingency")

    total_buildings = len(clipped_buildings)
    percentages = {
        key: (count / total_buildings) * 100 if total_buildings > 0 else 0
        for key, count in centroid_counts.items()
    }

    TP = centroid_counts["True Positive"]
    FP = centroid_counts["False Positive"]
    FN = centroid_counts["False Negative"]

    CSI = TP / (TP + FP + FN) if (TP + FP + FN) > 0 else 0
    FAR = FP / (TP + FP) if (TP + FP) > 0 else 0
    POD = TP / (TP + FN) if (TP + FN) > 0 else 0

    counts_data = {
        "Category": [
            "Candidate",
            "Benchmark",
            "False Positive",
            "False Negative",
            "True Positive",
            "CSI",
            "FAR",
            "POD",
        ],
        "Building Count": [
            centroid_counts["Candidate"],
            centroid_counts["Benchmark"],
            centroid_counts["False Positive"],
            centroid_counts["False Negative"],
            centroid_counts["True Positive"],
            f"{CSI:.3f}",
            f"{FAR:.3f}",
            f"{POD:.3f}",
        ],
    }

    counts_df = pd.DataFrame(counts_data)
    csv_file_path = os.path.join(
        save_dir, "EvaluationMetrics", f"BuildingCounts_{basename}.csv"
    )
    counts_df.to_csv(csv_file_path, index=False)

    third_raster_labels = ["False Positive", "False Negative", "True Positive"]
    third_raster_counts = [
        centroid_counts["False Positive"],
        centroid_counts["False Negative"],
        centroid_counts["True Positive"],
    ]

    fig = make_subplots(
        rows=1,
        cols=2,
        subplot_titles=(
            "Building Counts on Different FIMs",
            "Contingency Flooded Building Counts",
        ),
    )

    fig.add_trace(
        go.Bar(
            x=["Candidate"],
            y=[centroid_counts["Candidate"]],
            text=[f"{centroid_counts['Candidate']}"],
            textposition="auto",
            marker_color="#1c83eb",
            marker_line_color="black",
            marker_line_width=1,
            name=f"Candidate ({percentages['Candidate']:.2f}%)",
        ),
        row=1,
        col=1,
    )

    fig.add_trace(
        go.Bar(
            x=["Benchmark"],
            y=[centroid_counts["Benchmark"]],
            text=[f"{centroid_counts['Benchmark']}"],
            textposition="auto",
            marker_color="#a4490e",
            marker_line_color="black",
            marker_line_width=1,
            name=f"Benchmark ({percentages['Benchmark']:.2f}%)",
        ),
        row=1,
        col=1,
    )

    for i in range(len(third_raster_labels)):
        fig.add_trace(
            go.Bar(
                x=[third_raster_labels[i]],
                y=[third_raster_counts[i]],
                text=[f"{third_raster_counts[i]}"],
                textposition="auto",
                marker_color=["#ff5733", "#ffc300", "#28a745"][i],
                marker_line_color="black",
                marker_line_width=1,
                name=f"{third_raster_labels[i]} ({percentages[third_raster_labels[i]]:.2f}%)",
            ),
            row=1,
            col=2,
        )

    fig.update_layout(
        title="Flooded Building Counts",
        xaxis_title="Inundation Surface",
        yaxis_title="Flooded Building Counts",
        width=1100,
        height=400,
        plot_bgcolor="rgba(0, 0, 0, 0)",
        paper_bgcolor="rgba(0, 0, 0, 0)",
        showlegend=True,
        font=dict(family="Arial", size=18, color="black"),
    )

    plot_dir = os.path.join(save_dir, "FinalPlots")
    if not os.path.exists(plot_dir):
        os.makedirs(plot_dir)
    output_path = os.path.join(plot_dir, f"BuildingCounts_{basename}.png")
    fig.write_image(output_path, scale=500 / 96, engine="kaleido")
    print(f"Performance metrics chart is saved as PNG at {output_path}")
    fig.show()


def process_TIFF(
    tif_files, contingency_files, building_footprint, boundary, method_path
):
    benchmark_path = None
    candidate_path = []

    if len(tif_files) == 2:
        for tif_file in tif_files:
            if "benchmark" in tif_file.name.lower():
                benchmark_path = tif_file
            else:
                candidate_path.append(tif_file)

    elif len(tif_files) > 2:
        for tif_file in tif_files:
            if "benchmark" in tif_file.name.lower():
                benchmark_path = tif_file
                print(f"---Benchmark: {tif_file.name}---")
            else:
                candidate_path.append(tif_file)

    if benchmark_path and candidate_path:
        for candidate in candidate_path:

            matching_contingency_map = None
            candidate_base_name = candidate.stem.replace("_clipped", "")

            for contingency_file in contingency_files:
                if candidate_base_name in contingency_file.name:
                    matching_contingency_map = contingency_file
                    print(
                        f"Found matching contingency map for candidate {candidate.name}: {contingency_file.name}"
                    )
                    break

            if matching_contingency_map:
                print(
                    f"---FIM evaluation with Building Footprint starts for {candidate.name}---"
                )
                GetFloodedBuildingCountInfo(
                    building_footprint,
                    boundary,
                    benchmark_path,
                    candidate,
                    matching_contingency_map,
                    method_path,
                    candidate_base_name,
                )
            else:
                print(
                    f"No matching contingency map found for candidate {candidate.name}. Skipping..."
                )


def find_existing_footprint(out_dir):
    gpkg_files = list(Path(out_dir).glob("*.gpkg"))
    return gpkg_files[0] if gpkg_files else None

#Incase user defined individual shapefile for each case study
def detect_shapefile(folder):
        shapefile = None
        for ext in (".shp", ".gpkg", ".geojson", ".kml"):
            for file in os.listdir(folder):
                if file.lower().endswith(ext):
                    shapefile = os.path.join(folder, file)
                    print(f"Auto-detected shapefile: {shapefile}")
                    return shapefile
        return None
    
def EvaluationWithBuildingFootprint(
    main_dir,
    method_name,
    output_dir,
    country=None,
    building_footprint=None,
    shapefile_dir=None,
):
    tif_files_main = glob.glob(os.path.join(main_dir, "*.tif"))
    if tif_files_main:
        method_path = os.path.join(output_dir, os.path.basename(main_dir), method_name)
        for folder_name in os.listdir(method_path):
            if folder_name == "MaskedFIMwithBoundary":
                contingency_path = os.path.join(method_path, "ContingencyMaps")
                tif_files = list(
                    Path(os.path.join(method_path, folder_name)).glob("*.tif")
                )
                contingency_files = list(Path(contingency_path).glob("*.tif"))

                if shapefile_dir:
                    boundary = shapefile_dir
                elif os.path.exists(
                    os.path.join(method_path, "BoundaryforEvaluation")
                ):
                    boundary = os.path.join(
                        method_path, "BoundaryforEvaluation", "FIMEvaluatedExtent.shp"
                    )
                else:
                    boundary = detect_shapefile(main_dir)

                building_footprintMS = building_footprint
                if building_footprintMS is None:
                    import msfootprint as msf

                    out_dir = os.path.join(method_path, "BuildingFootprint")
                    if not os.path.exists(out_dir):
                        os.makedirs(out_dir)
                    EX_building_footprint = find_existing_footprint(out_dir)
                    if not EX_building_footprint:
                        boundary_dir = shapefile_dir if shapefile_dir else boundary

                        msf.BuildingFootprintwithISO(country, boundary_dir, out_dir)
                        building_footprintMS = os.path.join(
                            out_dir, f"building_footprint.gpkg"
                        )
                    else:
                        building_footprintMS = EX_building_footprint

                process_TIFF(
                    tif_files,
                    contingency_files,
                    building_footprintMS,
                    boundary,
                    method_path,
                )
    else:
        for folder in os.listdir(main_dir):
            folder_path = os.path.join(output_dir, folder)
            if os.path.isdir(folder_path):
                method_path = os.path.join(folder_path, method_name)
                for folder_name in os.listdir(method_path):
                    if folder_name == "MaskedFIMwithBoundary":
                        contingency_path = os.path.join(method_path, "ContingencyMaps")
                        tif_files = list(
                            Path(os.path.join(method_path, folder_name)).glob("*.tif")
                        )
                        contingency_files = list(Path(contingency_path).glob("*.tif"))

                        if shapefile_dir:
                            boundary = shapefile_dir
                        elif os.path.exists(
                            os.path.join(method_path, "BoundaryforEvaluation")
                        ):
                            boundary = os.path.join(
                                method_path, "BoundaryforEvaluation", "FIMEvaluatedExtent.shp"
                            )
                        else:
                            boundary = detect_shapefile(folder)

                        building_footprintMS = building_footprint
                        if building_footprintMS is None:
                            import msfootprint as msf

                            out_dir = os.path.join(method_path, "BuildingFootprint")
                            if not os.path.exists(out_dir):
                                os.makedirs(out_dir)
                            EX_building_footprint = find_existing_footprint(out_dir)
                            if not EX_building_footprint:
                                boundary_dir = (
                                    shapefile_dir if shapefile_dir else boundary
                                )
                                msf.BuildingFootprintwithISO(
                                    country, boundary_dir, out_dir
                                )
                                building_footprintMS = os.path.join(
                                    out_dir, f"building_footprint.gpkg"
                                )
                            else:
                                building_footprintMS = EX_building_footprint

                        process_TIFF(
                            tif_files,
                            contingency_files,
                            building_footprintMS,
                            boundary,
                            method_path,
                        )



from .evaluationFIM import EvaluateFIM
from .printcontingency import PrintContingencyMap
from .plotevaluationmetrics import PlotEvaluationMetrics
from .PWBs3 import get_PWB

__all__ = ["EvaluateFIM", "PrintContingencyMap", "PlotEvaluationMetrics", "get_PWB"]



import os
import numpy as np
from pathlib import Path
import geopandas as gpd
import rasterio
import shutil
import pandas as pd
from rasterio.warp import reproject, Resampling
from rasterio.io import MemoryFile
from rasterio import features
from rasterio.mask import mask

import warnings

warnings.filterwarnings("ignore", category=rasterio.errors.ShapeSkipWarning)

from .methods import AOI, smallest_extent, convex_hull, get_smallest_raster_path
from .metrics import evaluationmetrics
from .PWBs3 import get_PWB
from ..utilis import MakeFIMsUniform

# Function for the evalution of the model
def evaluateFIM(
    benchmark_path, candidate_paths, gdf, folder, method, output_dir, shapefile=None
):
    # Lists to store evaluation metrics
    csi_values = []
    TN_values = []
    FP_values = []
    FN_values = []
    TP_values = []
    TPR_values = []
    FNR_values = []
    Acc_values = []
    Prec_values = []
    sen_values = []
    F1_values = []
    POD_values = []
    FPR_values = []
    Merged = []
    Unique = []
    FAR_values = []
    Dice_values = []

    # Dynamically call the specified method
    method = globals().get(method)
    if method is None:
        raise ValueError(f"Method '{method}' is not defined.")

    # Save the smallest extent boundary and cliped FIMS
    save_dir = os.path.join(output_dir, os.path.basename(folder), f"{method.__name__}")
    os.makedirs(save_dir, exist_ok=True)

    # Get the smallest matched raster extent and make a boundary shapefile
    smallest_raster_path = get_smallest_raster_path(benchmark_path, *candidate_paths)

    #If method is AOI, and direct shapefile directory is not provided, then it will search for the shapefile in the folder
    if method.__name__ == "AOI":
        # If shapefile is not provided, search in the folder
        if shapefile is None:
            for ext in (".shp", ".gpkg", ".geojson", ".kml"):
                for file in os.listdir(folder):
                    if file.lower().endswith(ext):
                        shapefile = os.path.join(folder, file)
                        print(f"Auto-detected shapefile: {shapefile}")
                        break
                if shapefile:
                    break
            if shapefile is None:
                raise FileNotFoundError(
                    "No shapefile (.shp, .gpkg, .geojson, .kml) found in the folder and none provided. Either provide a shapefile directory or put shapefile inside folder directory."
                )

        # Run AOI with the found or provided shapefile
        bounding_geom = AOI(benchmark_path, shapefile, save_dir)

    else:
        print(f"--- {method.__name__} is processing ---")
        bounding_geom = method(smallest_raster_path, save_dir=save_dir)

    # Read and process benchmark raster
    with rasterio.open(benchmark_path) as src1:
        out_image1, out_transform1 = mask(
            src1, bounding_geom, crop=True, all_touched=True
        )
        benchmark_nodata = src1.nodata
        benchmark_crs = src1.crs
        b_profile = src1.profile
        out_image1[out_image1 == benchmark_nodata] = 0
        out_image1 = np.where(out_image1 > 0, 2, 0).astype(np.float32)
        gdf = gdf.to_crs(benchmark_crs)
        shapes1 = [
            geom for geom in gdf.geometry if geom is not None and not geom.is_empty
        ]
        mask1 = features.geometry_mask(
            shapes1,
            transform=out_transform1,
            invert=True,
            out_shape=out_image1.shape[1:],
        )
        extract_b = np.where(mask1, out_image1, 0)
        extract_b = np.where(extract_b > 0, 1, 0)
        idx_pwb = np.where(extract_b == 1)
        out_image1[idx_pwb] = 0

        benchmark_basename = os.path.basename(benchmark_path).split(".")[0]
        clipped_dir = os.path.join(save_dir, "MaskedFIMwithBoundary")
        if not os.path.exists(clipped_dir):
            os.makedirs(clipped_dir)

        clipped_benchmark = os.path.join(
            clipped_dir, f"{benchmark_basename}_clipped.tif"
        )
        b_profile.update(
            {
                "height": out_image1.shape[1],
                "width": out_image1.shape[2],
                "transform": out_transform1,
            }
        )

        with rasterio.open(clipped_benchmark, "w", **b_profile) as dst:
            dst.write(np.squeeze(out_image1), 1)

    def resize_image(
        source_image,
        source_transform,
        source_crs,
        target_crs,
        target_shape,
        target_transform,
    ):
        target_image = np.empty(target_shape, dtype=source_image.dtype)
        reproject(
            source=source_image,
            destination=target_image,
            src_transform=source_transform,
            dst_transform=target_transform,
            src_crs=source_crs,
            dst_crs=target_crs,
            resampling=Resampling.nearest,
        )
        return target_image

    # Process each candidate file
    for idx, candidate_path in enumerate(candidate_paths):
        base_name = os.path.splitext(os.path.basename(candidate_path))[0]
        with rasterio.open(candidate_path) as src2:
            candidate = src2.read(1)
            candidate_nodata = src2.nodata
            candidate_transform = src2.transform
            candidate_meta = src2.meta.copy()
            candidate_crs = src2.crs
            c_profile = src2.profile
            candidate[candidate == src2.nodata] = 0
            candidate = np.where(candidate > 0, 2, 1).astype(np.float32)
            with MemoryFile() as memfile:
                with memfile.open(**candidate_meta) as mem2:
                    mem2.write(candidate, 1)
                    dst_transform, width, height = (
                        rasterio.warp.calculate_default_transform(
                            mem2.crs,
                            benchmark_crs,
                            mem2.width,
                            mem2.height,
                            *mem2.bounds,
                        )
                    )
                    dst_meta = mem2.meta.copy()
                    dst_meta.update(
                        {
                            "crs": benchmark_crs,
                            "transform": dst_transform,
                            "width": width,
                            "height": height,
                        }
                    )

                    with MemoryFile() as memfile_reprojected:
                        with memfile_reprojected.open(**dst_meta) as mem2_reprojected:
                            for i in range(1, mem2.count + 1):
                                reproject(
                                    source=rasterio.band(mem2, i),
                                    destination=rasterio.band(mem2_reprojected, i),
                                    src_transform=mem2.transform,
                                    src_crs=mem2.crs,
                                    dst_transform=dst_transform,
                                    dst_crs=benchmark_crs,
                                    resampling=Resampling.nearest,
                                )
                            out_image2, out_transform2 = mask(
                                mem2_reprojected,
                                bounding_geom,
                                crop=True,
                                all_touched=True,
                            )
                            out_image2 = np.where(
                                out_image2 == candidate_nodata, 0, out_image2
                            )

                            # Save the clipped candidate raster
                            candidate_basename = os.path.basename(candidate_path).split(
                                "."
                            )[0]
                            clipped_candidate = os.path.join(
                                clipped_dir, f"{candidate_basename}_clipped.tif"
                            )
                            b_profile.update(
                                {
                                    "height": out_image1.shape[1],
                                    "width": out_image1.shape[2],
                                    "transform": out_transform1,
                                }
                            )
                            with rasterio.open(
                                clipped_candidate, "w", **b_profile
                            ) as dst:
                                dst.write(np.squeeze(out_image2), 1)

                            mask2 = features.geometry_mask(
                                shapes1,
                                transform=out_transform2,
                                invert=True,
                                out_shape=(out_image2.shape[1], out_image2.shape[2]),
                            )
                            extract_c = np.where(mask2, out_image2, 0)
                            extract_c = np.where(extract_c > 0, 1, 0)
                            idx_pwc = np.where(extract_c == 1)
                            out_image2[idx_pwc] = -1
                            out_image2_resized = resize_image(
                                out_image2,
                                out_transform2,
                                mem2_reprojected.crs,
                                benchmark_crs,
                                out_image1.shape,
                                out_transform1,
                            )
                            merged = out_image1 + out_image2_resized

            # Get Evaluation Metrics
            (
                unique_values,
                TN,
                FP,
                FN,
                TP,
                TPR,
                FNR,
                Acc,
                Prec,
                sen,
                CSI,
                F1_score,
                POD,
                FPR,
                merged,
                FAR,
                Dice,
            ) = evaluationmetrics(out_image1, out_image2_resized)

            # Append values to the lists
            csi_values.append(CSI)
            TN_values.append(TN)
            FP_values.append(FP)
            FN_values.append(FN)
            TP_values.append(TP)
            TPR_values.append(TPR)
            FNR_values.append(FNR)
            Acc_values.append(Acc)
            Prec_values.append(Prec)
            sen_values.append(sen)
            F1_values.append(F1_score)
            POD_values.append(POD)
            FPR_values.append(FPR)
            Merged.append(merged)
            Unique.append(unique_values)
            FAR_values.append(FAR)
            Dice_values.append(Dice)

    results = {
        "CSI_values": csi_values,
        "TN_values": TN_values,
        "FP_values": FP_values,
        "FN_values": FN_values,
        "TP_values": TP_values,
        "TPR_values": TPR_values,
        "FNR_values": FNR_values,
        "Acc_values": Acc_values,
        "Prec_values": Prec_values,
        "sen_values": sen_values,
        "F1_values": F1_values,
        "POD_values": POD_values,
        "FPR_values": FPR_values,
        # 'Merged': Merged,
        #  'Unique': Unique
        "FAR_values": FAR_values,
        "Dice_values": Dice_values,
    }
    for candidate_idx, candidate_path in enumerate(candidate_paths):
        candidate_BASENAME = os.path.splitext(os.path.basename(candidate_path))[0]
        merged_raster = Merged[candidate_idx]
        if merged_raster.ndim == 3:
            band = merged_raster.squeeze()
        elif merged_raster.ndim == 2:
            band = merged_raster
        else:
            raise ValueError(
                f"Unexpected number of dimensions in Merged[{candidate_idx}]."
            )

        # Construct the contingency file name dynamically
        contigency_dir = os.path.join(save_dir, "ContingencyMaps")
        os.makedirs(contigency_dir, exist_ok=True)
        output_filename = os.path.join(
            contigency_dir, f"ContingencyMAP_{candidate_BASENAME}.tif"
        )
        with rasterio.open(output_filename, "w", **b_profile) as dst:
            dst.write(band, 1)
            dst.transform = out_transform1
            dst.crs = benchmark_crs

    # Saving it into dataframe
    candidate_names = [
        os.path.splitext(os.path.basename(path))[0] for path in candidate_paths
    ]
    df = pd.DataFrame.from_dict(results, orient="index")
    df.columns = candidate_names
    df.reset_index(inplace=True)
    df.rename(columns={"index": "Metrics"}, inplace=True)

    # Save the DataFrame
    evaluationMetrics_DIR = os.path.join(save_dir, "EvaluationMetrics")
    os.makedirs(evaluationMetrics_DIR, exist_ok=True)

    csv_file = os.path.join(evaluationMetrics_DIR, "EvaluationMetrics.csv")
    df.to_csv(csv_file, index=False)
    print(f"Evaluation metrics saved to {csv_file}")
    return results



def EvaluateFIM(main_dir, method_name, output_dir, PWB_dir=None, shapefile_dir=None, target_crs=None, target_resolution=None):
    main_dir = Path(main_dir)
    # Read the permanent water bodies
    if PWB_dir is None:
        gdf = get_PWB()
    else:
        gdf = gpd.read_file(PWB_dir)

    def process_TIFF(tif_files, folder_dir):
        benchmark_path = None
        candidate_path = []

        if len(tif_files) == 2:
            for tif_file in tif_files:
                if "benchmark" in tif_file.name.lower() or "BM" in tif_file.name:
                    benchmark_path = tif_file
                else:
                    candidate_path.append(tif_file)

        elif len(tif_files) > 2:
            for tif_file in tif_files:
                if "benchmark" in tif_file.name.lower() or "BM" in tif_file.name:
                    benchmark_path = tif_file
                    print(f"---Benchmark: {tif_file.name}---")
                else:
                    candidate_path.append(tif_file)

        if benchmark_path and candidate_path:
            print(f"---Flood Inundation Evaluation of {folder_dir.name}---")
            Metrics = evaluateFIM(
                benchmark_path,
                candidate_path,
                gdf,
                folder_dir,
                method_name,
                output_dir,
                shapefile_dir,
            )
            print("\n", Metrics, "\n")
        else:
            print(
                f"Skipping {folder_dir.name} as it doesn't have a valid benchmark and candidate configuration."
            )

    # Check if main_dir directly contains tif files
    TIFFfiles_main_dir = list(main_dir.glob("*.tif"))
    if TIFFfiles_main_dir:
        MakeFIMsUniform(main_dir, target_crs=target_crs, target_resolution=target_resolution)

        #processing folder
        processing_folder = main_dir / "processing"
        TIFFfiles = list(processing_folder.glob("*.tif"))

        process_TIFF(TIFFfiles, main_dir)
        shutil.rmtree(processing_folder)
    else:
        for folder in main_dir.iterdir():
            if folder.is_dir():
                tif_files = list(folder.glob("*.tif"))
                
                if tif_files:
                    MakeFIMsUniform(folder, target_crs=target_crs, target_resolution=target_resolution)
                    #processing folder
                    processing_folder = folder / "processing"
                    TIFFfiles = list(processing_folder.glob("*.tif"))

                    process_TIFF(TIFFfiles, folder)
                    shutil.rmtree(processing_folder)
                else:
                    print(
                        f"Skipping {folder.name} as it doesn't contain any tif files."
                    )
